from typing import TypeVar

from rainbowadn.chain.states.activestateprotocol import ActiveStateProtocol
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.nullability.nullablereference import NullableReference

__all__ = ('MetaReductionStateProtocol',)

HeaderType = TypeVar('HeaderType')
StateType = TypeVar('StateType')


class MetaReductionStateProtocol(ActiveStateProtocol[HeaderType, StateType]):
    def verify(
            self,
            previous: NullableReference[StateType],
            header: HashPoint[HeaderType],
            state: HashPoint[StateType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(state, HashPoint)
        assert state == self.derive(previous, header)
        return True

    def _initial_state(self) -> HashPoint[StateType]:
        raise NotImplementedError

    def _derive(
            self,
            previous: HashPoint[StateType],
            header: HashPoint[HeaderType],
    ) -> HashPoint[StateType]:
        raise NotImplementedError

    def derive(
            self,
            previous: NullableReference[StateType],
            header: HashPoint[HeaderType],
    ) -> HashPoint[StateType]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        if previous.null():
            previous_state: HashPoint[StateType] = self._initial_state()
        else:
            previous_state: HashPoint[StateType] = previous.hashpoint()
        assert isinstance(previous_state, HashPoint)
        return self._derive(previous_state, header)
