from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.binary.binarytreecase import BinaryTreeCase
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol

__all__ = ('BinaryTreeAction',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')
ActionType = TypeVar('ActionType')


class BinaryTreeAction(Generic[ActiveKeyType, MetaDataType, TreeType, ActionType]):
    def on(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> ActionType:
        if (split := protocol.split(tree)) is None:
            return self.on_null(protocol, tree)
        else:
            return self.on_split(BinaryTreeCase(protocol, split, tree))

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> ActionType:
        raise NotImplementedError

    def on_split(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> ActionType:
        raise NotImplementedError

    def per(
            self: 'BinaryTreeAction[ActiveKeyType, MetaDataType, TreeType, ActionType]',
            _protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType]
    ) -> 'BinaryTreeAction[ActiveKeyType, MetaDataType, TreeType, ActionType]':
        return self
