from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.binary.actions.binarytreeaction import BinaryTreeAction
from rainbowadn.data.collection.trees.binary.actions.compareaction import CompareAction
from rainbowadn.data.collection.trees.binary.binarytreecase import BinaryTreeCase
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.data.collection.trees.comparison.comparator import Equal, Replace

__all__ = ('AddAction', 'RemoveAction', 'ContainsAction',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class AddAction(
    CompareAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def on_equal(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> TreeType:
        if isinstance(equal, Replace):
            return case.protocol.tree(case.split.treel, case.split.treer, self.key)
        else:
            raise TypeError

    def on_left(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return case.protocol.tree(self.on(case.protocol, case.split.treel), case.split.treer, case.split.key)

    def on_right(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return case.protocol.tree(case.split.treel, self.on(case.protocol, case.split.treer), case.split.key)

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> TreeType:
        return protocol.tree(tree, tree, self.key)


class MergeAction(
    BinaryTreeAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(self, treer: TreeType):
        self.treer = treer

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> TreeType:
        return self.treer

    def on_split(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        protocol = case.protocol
        split = case.split
        return protocol.tree(
            split.treel,
            MergeAction(self.treer).on(protocol, split.treer),
            split.key
        )


class RemoveAction(
    CompareAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def on_equal(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> TreeType:
        return MergeAction(case.split.treer).on(case.protocol, case.split.treel)

    def on_left(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return case.protocol.tree(self.on(case.protocol, case.split.treel), case.split.treer, case.split.key)

    def on_right(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return case.protocol.tree(case.split.treel, self.on(case.protocol, case.split.treer), case.split.key)

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> TreeType:
        return tree


class ContainsAction(
    CompareAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        bool
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def on_equal(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> bool:
        return True

    def on_left(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> bool:
        return self.on(case.protocol, case.split.treel)

    def on_right(
            self,
            case: BinaryTreeCase[ActiveKeyType, MetaDataType, TreeType]
    ) -> bool:
        return self.on(case.protocol, case.split.treer)

    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ) -> bool:
        return False
